from moz_sql_parser.keywords import *
UNION_DISTINCT = Group(UNION + DISTINCT).set_parser_name("union_distinct")

reserved = [
    ALL,
    AND,
    AS,
    ASC,
    BETWEEN,
    BY,
    CASE,
    CAST,
    COLLATE,
    CROSS_JOIN,
    CROSS,
    DESC,
    DISTINCT,
    ELSE,
    END,
    FALSE,
    FROM,
    FULL_JOIN,
    FULL_OUTER_JOIN,
    FULL,
    GROUP_BY,
    GROUP,
    HAVING,
    IN,
    INNER_JOIN,
    INNER,
    INTERVAL,
    IS_NOT,
    IS,
    JOIN,
    LEFT_JOIN,
    LEFT_OUTER_JOIN,
    LEFT,
    LIKE,
    LIMIT,
    NOCASE,
    NOT_BETWEEN,
    NOT_IN,
    NOT_LIKE,
    NOT_RLIKE,
    NOT,
    NULL,
    OFFSET,
    ON,
    OR,
    ORDER_BY,
    ORDER,
    OUTER,
    OVER,
    PARTITION_BY,
    PARTITION,
    RIGHT_JOIN,
    RIGHT_OUTER_JOIN,
    RIGHT,
    RLIKE,
    SELECT_DISTINCT,
    SELECT,
    THEN,
    TRUE,
    UNION_DISTINCT,
    UNION_ALL,
    UNION,
    USING,
    WHEN,
    WHERE,
    WITH,
    WITHIN_GROUP,
    WITHIN,
]

unions = [UNION_DISTINCT, UNION_ALL, UNION]

min_keys = {
    "cast": 2,
    "mul": 2,
    "div": 2,
    "mod": 2,
    "neg": 2,
    "add": 2,
    "sub": 2,
    "case": 2,
    "binary_not": 2,
    "binary_and": 2,
    "binary_or": 2,
    "gte": 2,
    "lte": 2,
    "lt": 2,
    "gt": 2,
    "eq": 2,
    "neq": 2,
    "between": 2,
    "not_between": 2,
    "in": 2,
    "nin": 2,
    "is": 2,
    "like": 2,
    "not_like":2,
    "rlike":2,
    "not_rlike": 2,
    "similar_to": 2,
    "not_similar_to": 2,
}


from .dialects import load_dialect_module
from .clean import strip_comments, unwrap_variables, ENCODING_CHARACTER
from moz_sql_parser.formatting import Formatter as MozFormatter, escape


def format_query(query_ast, sql_dialect) -> str:
    """

    :param query_ast:
    :param sql_dialect:
    :return:
    """
    dialect = load_dialect_module(dialect_name=sql_dialect)
    formatted_query = dialect.format_query(query_ast=query_ast)

    formatted_query = unwrap_variables(
        query=formatted_query, encoding_character=ENCODING_CHARACTER, skip_characters=len(ENCODING_CHARACTER)
    )
    formatted_query = unwrap_variables(
        query=formatted_query, encoding_character='"',
        skip_characters=len(ENCODING_CHARACTER)
    )
    return formatted_query


class BaseFormatter(MozFormatter):
    """"""

    def _union(self, json, wrap=True):
        return "(" + self.union(json) + ")"

    def _union_all(self, json, wrap=True):
        return "(" + self.union_all(json) + ")"

    def union_distinct(self, json):
        return " UNION DISTINCT ".join(self.query(query) for query in json)

    def _union_distinct(self, json):
        return "(" + self.union_distinct(json=json) + ')'

    def _interval(self, json):
        return 'INTERVAL {0} {1}'.format(*json)

    def _distinct(self, json):
        if isinstance(json, list):
            return "DISTINCT " + self.dispatch(json=json)
        else:
            return "DISTINCT(" + self.dispatch(json=json) +')'

    def format(self, json):
        if "union" in json:
            return self.union(json["union"])
        elif "union_all" in json:
            return self.union_all(json["union_all"])
        elif "union_distinct" in json:
            return self.union_distinct(json["union_distinct"])
        else:
            return self.query(json)

    def _in(self, json):
        valid = self.dispatch(json[1])
        # `(10, 11, 12)` does not get parsed as literal, so it's formatted as
        # `10, 11, 12`. This fixes it.
        if not valid.startswith("("):
            valid = "({0})".format(valid)

        return "{0} IN {1}".format(self.dispatch(json[0]), valid)

    def _cast(self, json):
        _type = json[1]
        if isinstance(_type, dict):
            _type = list(_type.keys())[0]
        return "cast({0} AS {1})".format(json[0], _type)



def format_query(query_ast: dict) -> str:
    from .format import BaseFormatter
    return BaseFormatter().format(query_ast)


def parse_query(query: str) -> dict:
    from .parse import Parser
    return Parser().parse(query)



import re

ENCODING_CHARACTER = "'"


def clean_query(query, sql_dialect) -> str:
    """

    :param query:
    :param sql_dialect:
    :return:
    """
    cleaned_query = strip_comments(query=query)
    cleaned_query = pad_comparisons(query=cleaned_query)
    cleaned_query = clean_spaces(query=cleaned_query)
    return cleaned_query.lower()


def pad_comparisons(query):
    import re
    cleaned_query = re.sub(
        '([\s\(][\d\w\.\_\{\}]*)(<=|>=|<>|==|!=|<|>|=|\|\||::|\+|-|/|\*|~)([\d\w\.\_\{\}]*[\)\s]*)', r' \1 \2 \3',
        query)
    return cleaned_query


def clean_spaces(query) -> str:
    """

    :param query:
    :return:
    """
    cleaned_query = re.sub('\s+', ' ', query)
    cleaned_query = re.sub('\(\s', '(', cleaned_query)
    cleaned_query = re.sub('\s\)', ')', cleaned_query)
    return cleaned_query.strip()


def strip_comments(query) -> str:
    """

    :param query:
    :return:
    """
    return re.sub('(([\s-]+-).+\n)|(([\s\#]+\#).+\n)', ' ', query)


def wrap_variables(query, encoding_character='', wrap_character="'", encoded=False):
    """

    :param query:
    :param encoding_character:
    :param wrap_character:
    :param encoded:
    :param skip_characters:
    :param feature_name:
    :return:
    """
    if encoded and encoding_character:
        encoding_character = '\\' + encoding_character
    regex = '(' + encoding_character + '[\w\d\_]*\{\S+\}[\w\d\_]*' + encoding_character + ')'
    variables = set()
    query = re.sub(regex, r"{0}\1{0}".format(wrap_character), query)
    for match in re.findall('(' + wrap_character + '[\w\d\_]*\{\S+\}[\w\d\_]*' + wrap_character + ')', query):
        new_name = match
        variables.add(new_name)

    return query, variables


def unwrap_variables(query, encoding_character='', encoded=False, skip_characters=0):
    if encoded and encoding_character:
        encoding_character = '\\' + encoding_character
    regex = '(' + encoding_character + '[\w\d\_]*\{\S+\}[\w\d\_]*' + encoding_character + ')'
    for match in re.findall(regex, query):
        new_name = match
        new_string = new_name[skip_characters:len(new_name) - skip_characters]
        query = query.replace(
            match,
            new_string
        )
    return query


def prune_ast(json, parent=None):
    """ Recursive function to remove partial or nulled values from the AST

    :param json: json query AST
    :param parent: the parent key of the ast
    :return: cleaned query AST
    """
    pruned = type(json)()
    if isinstance(json, dict):
        for k, v in json.items():
            child = prune_ast(json=v, parent=k)
            if child:
                pruned[k] = child
        pruned = validate_tree(k=parent, json=pruned)
    elif isinstance(json, list):
        for v in json:
            child = prune_ast(json=v, parent=None)
            if child:
                pruned.append(child)
        pruned = validate_tree(k=parent, json=pruned)
    else:
        if json is not None:
            pruned = json
    return pruned


def validate_tree(k, json):
    """ Validates whether the query ast has the required number of values

    :param k: the query key
    :param json: the query AST
    :return: None if invalid and the query_ast if valid.
    """
    from camtono.parser.parse import min_keys
    if k is None or len(json) >= min_keys.get(k, 1):
        return json
    else:
        return None


def set_value(val, **kwargs):
    """ Convenience function to set value for set_tree_value

    :param val: value to return
    :param kwargs: all other values
    :return: the value provided
    """
    return val


def set_tree_value(json, locations, replace_func=set_value, val=None, reverse_index=None, target=False):
    """ Set the

    :param json: dictionary
    :param locations: dictionary of the paths containing the location of a target value
    :param replace_func: function to apply when setting value receives val and json
    :param val: value to set
    :param reverse_index: location from the end of the tree where the replacement function should be applied
    :param target: boolean value to force value application
    :return: dictionary with the newly assigned values.
    """
    if locations and not target:
        current_index = None
        for k, v in locations.items():
            if k.isdigit():
                k = int(k)
            v, index = set_tree_value(json=json[k], locations=v, val=val, replace_func=replace_func,
                                      reverse_index=reverse_index)
            current_index = index + 1
            if reverse_index is not None and current_index == reverse_index:
                v, index = set_tree_value(json=json[k], locations=v, val=val, replace_func=replace_func, target=True,
                                          reverse_index=reverse_index)
            json[k] = v
        return json, current_index
    else:
        index = 0
        return replace_func(json=json, val=val), index



def generate_feature(name, query, sql_dialect) -> dict:
    """

    :param name:
    :param query:
    :param sql_dialect:
    :return:
    """
    from .parse import parse_query
    ast, query_inputs, query_outputs, dependencies = parse_query(query=query, sql_dialect=sql_dialect,
                                                                 feature_name=name)
    feature = dict(
        name=name,
        inputs=query_inputs,
        outputs=query_outputs,
        dependencies=dependencies,
        query_ast=ast,
        allowed_dialects=get_allowed_dialects(query_ast=ast)
    )
    return feature


def get_allowed_dialects(query_ast):
    """

    :param query_ast:
    :return:
    """
    from .format import format_query
    from .dialects import list_dialects

    allowed_dialects = []
    for dialect in list_dialects():
        try:
            format_query(query_ast=query_ast, sql_dialect=dialect)
        except Exception:
            pass
        else:
            allowed_dialects.append(dialect)
    return allowed_dialects



def load_dialect_module(dialect_name):
    from importlib import import_module
    if dialect_name != 'standard':
        dialect = import_module('camtono.dialects.{}'.format(dialect_name))
    else:
        dialect = import_module('camtono.parser.{}'.format(dialect_name))
    return dialect


def list_dialects():
    """

    :return:
    """
    import pkgutil
    try:
        import camtono.dialects
    except ModuleNotFoundError:
        pass
    else:
        for finder, modname, ispkg in pkgutil.iter_modules(camtono.dialects.__path__):
            yield modname
    finally:
        yield "standard"
